// $Id: CasClient.java,v 1.37 2007-12-05 08:48:30+09 naito Exp $

package jp.ac.nagoya_u.cas.client;

import java.io.*;
import java.util.*;
import java.net.URLEncoder;

import javax.servlet.*;
import javax.servlet.http.*;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * CAS client module for JAVA (Nagoya University)
 * <BR>
 * 実際の利用法： 
 * サーブレット内の認証を行うべき場所に以下のコードを挿入.
 * <pre>
 * CasClient cas = new CasClient() ;
 * cas.setServletURL(CAS_SERVICE_URL) ;
 * cas.setLoginURL(CAS_LOGIN_URL) ;
 * if (!cas.casPerform(request, response)) return ;
 * Map r = cas.getResult() ;
 * </pre>
 * 
 * または
 *
 * <pre>
 * CasClient cas = new CasClient(CAS_SERVICE_URL, CAS_LOGIN_URL) ;
 * if (!cas.casPerform(request, response)) return ;
 * Map r = cas.getResult() ;
 * </pre>
 *
 * このプログラム断片が終了すると, <code> r </code> にはユーザ情報が書き込まれているので, 
 * <pre>
 * (String) r.get("uid") ;
 * </pre>
 * とすれば, <code> uid </code> 属性値を取得することができます.
 * <BR>
 * このプログラムを利用する際には, 
 * CAS_VALIDATE_URL, CAS_ERROR_URL 
 * を適切なものに書き換える必要があります.
 *
 * @version 1.2.4.2
 * @author  Hisashi NAITO (naito@math.nagoya-u.ac.jp)
 */
public class CasClient {
    /** */
    private static final String VALIDATE_URL = "serviceValidate" ;
    private static final String ERROR_URL    = "backService.jsp" ;
    private static final String CAS_TICKET   = "ticket" ;
    ///////////////////////////////////////////////////////////////
    private String casLoginURL    ;
    private String casServiceURL  ;
    private String casValidateURL ;
    private String casErrorURL ;
    private String errorCode    ;
    private String errorMessage ;
    private String xmlResponse  ; 
    private Map r ;
    private boolean useSSL = true ;
    private String exceptingParameter = "username,password,lt,doubly,service" ;
    private String clientEncoding = "UTF-8" ;
    private boolean isURIEncoded = false ;

    /** Log instance for logging events, info, warnings, errors, etc. */
    private final Log log = LogFactory.getLog(this.getClass());

    public CasClient() 
    {
	super() ;
    }

    /**
     * Construct CasClient instance
     *
     * @param casServiceURL URL of the application.
     * @param casLoginURL   URL of the login servlet of CAS
     */
    public CasClient(String casServiceURL, String casLoginURL) 
    {
	this.casServiceURL  = casServiceURL ;
	this.casLoginURL    = casLoginURL ;
	this.casValidateURL = setCasValidateURL(this.casLoginURL) ;
	this.casErrorURL    = setCasErrorURL(this.casLoginURL) ;
	if (log.isDebugEnabled()) {
	    log.debug("CAS Client Constructor") ;
	    log.debug("CAS Login URL    = " + this.casLoginURL) ;
	    log.debug("CAS Sevice URL   = " + this.casServiceURL) ;
	    log.debug("CAS Validate URL = " + this.casValidateURL) ;
	    log.debug("CAS Error URL    = " + this.casErrorURL) ;
	}
    }

    /**
     * Construct CasClient instance.
     * Use this constructor, if the validation servlet of CAS does not same FQDN of the login servlet of CAS
     *
     * @since 1.2.4.2
     *
     * @param casServiceURL    URL of the application.
     * @param casLoginURL      URL of the login servlet of CAS
     * @param casValidateURL   URL of the validation servlet of CAS
     */
    public CasClient(String casServiceURL, String casLoginURL, String casValidateURL) 
    {
	this.casServiceURL  = casServiceURL ;
	this.casLoginURL    = casLoginURL ;
	this.casValidateURL = casValidateURL ;
	this.casErrorURL    = setCasErrorURL(this.casLoginURL) ;
	if (log.isDebugEnabled()) {
	    log.debug("CAS Client Constructor") ;
	    log.debug("CAS Login URL    = " + this.casLoginURL) ;
	    log.debug("CAS Sevice URL   = " + this.casServiceURL) ;
	    log.debug("CAS Validate URL = " + this.casValidateURL) ;
	    log.debug("CAS Error URL    = " + this.casErrorURL) ;
	}
    }

    /**
     * perform cas client
     * @return TRUE if validation success, otherwise FALSE.
     */
    synchronized public boolean casPerform(ServletContext app, HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException 
    {
	return casPerform(request, response) ;
    }

    /**
     * perform cas client
     * @return TRUE if validation success, otherwise FALSE.
     */
    synchronized public boolean casPerform(String casLoginURL, HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException 
    {
	this.casLoginURL    = casLoginURL ;
	this.casValidateURL = setCasValidateURL(this.casLoginURL) ;
	this.casErrorURL    = setCasErrorURL(this.casLoginURL) ;
	return casPerform(request, response) ;
    }

    /**
     * perform cas client
     * @return TRUE if validation success, otherwise FALSE.
     */
    synchronized public boolean casPerform(String casLoginURL, String casValidateURL, HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException 
    {
	this.casLoginURL    = casLoginURL ;
	this.casValidateURL = casValidateURL + VALIDATE_URL ;
	this.casErrorURL    = setCasErrorURL(this.casLoginURL) ;
	return casPerform(request, response) ;
    }

    /**
     * perform cas client
     * @return TRUE if validation success, otherwise FALSE.
     */
    synchronized public boolean casPerform(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException 
    {
	String requestMethod = request.getMethod().toUpperCase() ;
	if ((!this.useSSL)&&(this.casValidateURL.startsWith("https:"))) {
	    this.casValidateURL = this.casValidateURL.replaceFirst("https:", "http:") ;
	}
  	if (log.isDebugEnabled()) {
	    log.debug("CAS Login URL    = " + this.casLoginURL) ;
	    log.debug("CAS Sevice URL   = " + this.casServiceURL) ;
	    log.debug("CAS Validate URL = " + this.casValidateURL) ;
	    log.debug("CAS Error URL    = " + this.casErrorURL) ;
	    log.debug("Request Method   = " + requestMethod) ;
  	}

	/* if ticket is null, sendRedirect and return false */
	String	ticket = request.getParameter(CAS_TICKET) ;
	if (ticket == null) {
	    // If method is GET, then redirect, else if method is POST, then generate JavaScript Page
	    log.debug("Ticket is null. response") ;
	    if (requestMethod.equals("GET")) response.sendRedirect(makeResponseString(request)) ;
	    else {
		PrintWriter clientWriter = response.getWriter() ;
		clientWriter.println(makePostHTMLString(request)) ;
		clientWriter.flush() ;
		clientWriter.close() ;
	    }
	    return false ;
	}
	log.debug("Ticket is non null. Validate ticket.") ;
	/* Ticket is non-null, Validate the ticket */
	ServiceTicketValidator sv = new ServiceTicketValidator() ;
	sv.setUseSSL(this.useSSL) ;
	sv.setCasValidateUrl(casValidateURL) ;
	sv.setServiceTicket(request.getParameter(CAS_TICKET)) ;
	sv.setService(this.casServiceURL) ;
	try {
	    log.debug("Try to Validate") ;
	    sv.validate() ;
	} catch (Exception e) {
	    request.setAttribute("errorMsg", "Could not obtain response: " + e) ;
	    log.debug("Validate Fail: " + e) ;
	    return false ;
	}
	log.debug("Success to Validate") ;

	xmlResponse  = sv.getResponse() ;
	r = sv.getResult() ;

	/* If Authenticaion is failure, */
	/* and if errorCode is INVALID_TICKET, */
	/* then redirect to CAS LOGIN */
	/* else (ACCESS DENIED), then redirect to CAS ERROR */
	if (!sv.isAuthenticationSuccesful()) { 
	    errorCode    = sv.getErrorCode() ;
	    log.debug("Error: errorCode = " + errorCode) ;
	    //*** Ticket is INVALID, get new ticket ***//
	    if (errorCode.equals("INVALID_TICKET")) {
		// If method is GET, then redirect, else if method is POST, then generate JavaScript Page
		log.debug("Ticket is INVALID_TICKET. response") ;
		if (requestMethod.equals("GET")) response.sendRedirect(makeResponseString(request)) ;
		else {
		    PrintWriter clientWriter = response.getWriter() ;
		    clientWriter.println(makePostHTMLString(request)) ;
		    clientWriter.flush() ;
		    clientWriter.close() ;
		}
		return false ;
	    }
	    else {
		//*** ACCESS Denied ***//
		log.debug("Redirect to " + this.casErrorURL) ;
		response.sendRedirect(this.casErrorURL) ;
		return false ;
	    }
	}
	log.debug("Finish to validate") ;
	return true ;
    }

    private String makeResponseString (HttpServletRequest request) {
	StringBuffer buffer = new StringBuffer() ;
	buffer.append(this.casLoginURL) ; buffer.append("?service=") ; buffer.append(this.casServiceURL) ;
	Enumeration e = request.getParameterNames() ;
	synchronized (e) {
	    while(e.hasMoreElements()) {
		String k = (String)e.nextElement() ;
		// Servlet 2.3 (tomcat 5.x) では, characterEncoding は URL パラメータに対しては適用されない可能性がある.
		// そのため, characterEncoding に従って, コード変換する必要がある.
		try {
		    String requestParameter = new String(request.getParameter(k).getBytes("iso-8859-1"), this.clientEncoding) ;
		    buffer.append("&") ; buffer.append(k) ; buffer.append("=") ; 
		    // buffer.append(URLEncoder.encode(escapeXml(requestParameter), this.clientEncoding)) ;
		    buffer.append(URLEncoder.encode(requestParameter, this.clientEncoding)) ;
		    // log.debug("Parameter " + k + "=" + requestParameter) ;
		    // log.debug(this.clientEncoding + " Encoded Parameter " + k + "=" + URLEncoder.encode(requestParameter, this.clientEncoding)) ;
		} catch (UnsupportedEncodingException ue) {
		    log.error(ue) ;
		}
	    }
	}
	if (!this.clientEncoding.equalsIgnoreCase("UTF-8")) {
	    buffer.append("&") ;
	    buffer.append("casClientEncoding=" + this.clientEncoding) ;
	}
	log.debug("response String = " + buffer.toString()) ;
	return buffer.toString() ;
    }


    /**
     * Post の時の Redirect Java Script を生成
     *
     * @return Java Script Code
     */
    private String makePostHTMLString(HttpServletRequest request) {
	StringBuffer requestParameterStringBuffer       = new StringBuffer() ;
	StringBuffer requestSubmitParameterStringBuffer = new StringBuffer() ;
	StringBuffer requestSubmitParameterStringBufferOfNoScript = new StringBuffer() ;
	/* ************** */
	Map exceptingParameterMap = new HashMap() ;
	String[] list = exceptingParameter.split(",") ;
	for(int i=0;i<list.length;i++) exceptingParameterMap.put(list[i], "") ;
	exceptingParameterMap.put("service","") ;
	/* ************** */
	Enumeration e = request.getParameterNames() ;
	synchronized (e) {
	    while(e.hasMoreElements()) {
		String k = (String)e.nextElement() ;
		if (exceptingParameterMap.get(k) == null) {
		    try {
			if (k.equals("submit")) {
			    requestSubmitParameterStringBuffer.append("<input type=hidden name=\"" + k + "\" value=\"" + URLEncoder.encode(request.getParameter(k),this.clientEncoding) + "\">\n") ;
			    requestSubmitParameterStringBufferOfNoScript.append("<input type=submit name=\"" + k + "\" value=\"" + URLEncoder.encode(request.getParameter(k),this.clientEncoding) + "\">\n") ;
			}
			else {
			    requestParameterStringBuffer.append("<input type=hidden name=\"" + k + "\" value=\"" + URLEncoder.encode(request.getParameter(k),this.clientEncoding) + "\">\n") ;
			    // log.debug(this.clientEncoding + " Encodeed Parameter " + k + "=" + URLEncoder.encode(request.getParameter(k), this.clientEncoding)) ;
			}
		    } catch (Exception ee) {
		    }
		}
	    }
	}
	/* ************** */
	StringBuffer buffer = new StringBuffer() ;
	buffer.append("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n") ;
	buffer.append("<html><head>\n") ;
	buffer.append("<title>Central Authentication and Authorization Service</title>\n") ;
	buffer.append("<meta http-equiv=\"Content-Type\" content=\"text/html;charset=" + this.clientEncoding + "\">\n") ;
	buffer.append("<meta http-equiv=\"content-script-type\" content=\"text/javascript\">\n") ;
	buffer.append("</head>\n") ;
	buffer.append("<body bgcolor=\"#ffffff\" link=\"#000000\" alink=\"#000000\" vlink=\"#000000\" onLoad=\"document.form.submit()\">\n") ;
	buffer.append("<noscript>Push&nbsp;</noscript>\n") ;
	buffer.append("<form name=\"form\" action=\"" + this.casLoginURL + "?service=" + this.casServiceURL + "\" method=POST>\n") ;
	buffer.append(requestParameterStringBuffer.toString()) ;
//  	if (!this.clientEncoding.equalsIgnoreCase("UTF-8")) {
//  	    buffer.append("<input type=hidden name=\"casClientEncoding\" value=\"" + this.clientEncoding + "\">\n") ;
//  	}
	buffer.append("<script>\n") ;	
	buffer.append(requestSubmitParameterStringBuffer.toString()) ;
	buffer.append("</script>\n") ;
	buffer.append("<noscript>\n") ;	
	buffer.append(requestSubmitParameterStringBufferOfNoScript.toString()) ;
	buffer.append("to access the service you requested.</noscript>\n") ;
	buffer.append("</form>\n") ;
	buffer.append("Central Authentication and Authorization Service Java Client Module \n") ;
	buffer.append(CasClientVersion.getVersion()) ;
	buffer.append("</body></html>\n") ;
	return buffer.toString() ;
    }

    private String escapeXml(String str) {
	return str.replaceAll("<", "&lt;").replaceAll(">", "&gt;").replaceAll("\"", "&quot;") ;
    }

    //********* get Methods *****************/
    /**
     * @return result Map
     */
    public Map getResult() 
    {
	return r ;
    }

    /**
     * @return entire CAS response 
     */
    public String getResponse()
    {
	return xmlResponse  ; 
    }

    //********* set Methods *****************/
    /**
     * set casServiceURL
     */
    public void setServiceURL(String casServiceURL) 
    {
	this.casServiceURL = casServiceURL ;
	return ;
    }

    /**
     * set casLoginURL
     */
    public void setLoginURL(String casLoginURL) 
    {
	this.casLoginURL = casLoginURL ;
	this.casValidateURL = setCasValidateURL(this.casLoginURL) ;
	this.casErrorURL    = setCasErrorURL(this.casLoginURL) ;
	return ;
    }

    public String setCasValidateURL(String casLoginURL) 
    {
	if (casLoginURL.endsWith("/")) return casLoginURL + VALIDATE_URL ;
	int lastIndex = casLoginURL.lastIndexOf("/") ;
	return casLoginURL.substring(0,lastIndex) + "/" + VALIDATE_URL ;
    }

    public String setCasErrorURL(String casLoginURL) 
    {
	if (casLoginURL.endsWith("/")) return casLoginURL + ERROR_URL ;
	int lastIndex = casLoginURL.lastIndexOf("/") ;
	return casLoginURL.substring(0,lastIndex) + "/" + ERROR_URL ;
    }

    public void setUseSSL(boolean useSSL) 
    {
	this.useSSL = useSSL ;
    }

    public void setClientEncoding(String clientEncoding) 
    {
	this.clientEncoding = clientEncoding ;
    }

    public void setIsURIEncoded(boolean isURIEncoded) 
    {
	this.isURIEncoded = isURIEncoded ;
    }

    public void setExceptingParameter(String exceptingParameter) 
    {		
	StringBuffer buffer = new StringBuffer() ;
	buffer.append("this.exceptingParameter") ;
	buffer.append(",") ;
	buffer.append(exceptingParameter) ;
	this.exceptingParameter = buffer.toString() ;
	return ;
    }

}
